﻿using System;
using System.Collections.Generic;
using UnityEngine;
using UnityEngine.Video;
using UnityEngine.UI;
using System.Linq;

namespace VamilaCode
{
    public class RenderVideo : MVRScript
    {
        #region Storable variables        
        public JSONStorableStringChooser videoObject;
        public JSONStorableString videoClipPath;        
        public JSONStorableBool isLooping;
        public JSONStorableBool isPlaying;        
        public JSONStorableBool audioSpatialize;
        public JSONStorableFloat audioVolume;        
        public JSONStorableFloat audioSpatialBlend;
        public JSONStorableFloat audioSpread;
        public JSONStorableFloat audioMinDistance;
        public JSONStorableFloat audioMaxDistance;       
        #endregion

        #region Private variables
        // UI Button instance
        private UIDynamicButton chooseVideoBtn;
        // VideoPlayer component instance
        private VideoPlayer videoPlayer;
        // AudioSource component instance
        private AudioSource audioSource;        
        // Default path for stored videos
        private readonly string DefaultVideoPath = @"Saves\\Videos\\";       
        /// <summary>
        /// Key = supported atom type, Value = object to render video.
        /// </summary>
        private Dictionary<string, List<string>> supportedAtoms = new Dictionary<string, List<string>>()
        {
            {
                "CyberpunkLaptop", new List<string>()
                {
                    "Laptop Screen Image"
                }
            },
            {
                "CyberpunkComputer", new List<string>()
                {
                    "Screen 1 Image",
                    "Screen 1 Image Solo",
                    "Screen 2 Image",
                    "Screen 2 Image Solo",
                }
            },
            {
                "CyberpunkControlScreen", new List<string>()
                {
                    "Screen Image Center",
                    "Screen Image",
                    "Screen Image Right",                    
                    "Screen Image Left"
                }
            },
            {
                "CyberpunkTablet", new List<string>()
                {
                    "Screen Right Image",
                    "Screen Left Image",
                    "Screen Portrait Image"
                }
            },
            {
                "AptSmartTV", new List<string>()
                {
                    "Image"
                }
            },
            {
                "DreamHomeTV", new List<string>()
                {
                    "Image"
                }
            },
            {
                "Image", new List<string>()
                {
                    "Image"
                }
            },
            {
                "ImagePanelEmissive", new List<string>()
                {
                    "Image"
                }
            },
            {
                "ImagePanelTransparent", new List<string>()
                {
                    "Image"
                }
            },
            {
                "ImagePanelTransparentEmissive", new List<string>()
                {
                    "Image"
                }
            }
        };
        #endregion

        #region Utils
        /// <summary>
        /// Find child GameObject.
        /// </summary>
        /// <param name="parentObjectName">Parent GameObject name.</param>
        /// <param name="childObjectName">Child GameObject name.</param>
        /// <returns>Child GameObject</returns>
        private GameObject GetImageGameObject(GameObject parentObjectName, string childObjectName)
        {
            if (parentObjectName != null)
            {
                Transform[] tforms = parentObjectName.transform.GetComponentsInChildren<Transform>(true);
                foreach (Transform tform in tforms)
                    if (tform.gameObject.name == childObjectName)
                        if (tform.gameObject.GetComponent<ImageControl>()!=null)
                            return tform.gameObject;
            }

            return null;
        }

        /// <summary>
        /// Creates script caption text with description.
        /// </summary>
        /// <param name="caption">Script caption.</param>        
        private UIDynamicTextField CreateScriptCaptionUI(JSONStorableString caption)
        {
            var result = CreateTextField(caption);
            result.height = 120f;
            result.textColor = new Color(255, 255, 255, 0.6f);
            result.backgroundColor = new Color(0, 0, 0, 0.6f);

            return result;
        }

        /// <summary>
        /// Create group of fields caption.
        /// </summary>
        /// <param name="caption">Group caption.</param>
        /// <returns></returns>
        private UIDynamicTextField CreateGroupCaptionUI(JSONStorableString caption, bool right = false)
        {
            var result = CreateTextField(caption, right);
            result.height = 40f;
            var layout = result.GetComponent<LayoutElement>();
            layout.minHeight = layout.preferredHeight = 40f;            
            result.textColor = new Color(255, 255, 255, 0.6f);
            result.backgroundColor = new Color(0, 0, 0, 0);

            return result;
        }
        #endregion

        #region Callbacks
        /// <summary>
        /// Play/pause video clip.
        /// </summary>
        /// <param name="value">true - play, false - pause</param>
        private void PlayToggle(bool value)
        {
            if (string.IsNullOrEmpty(videoClipPath.val)) return;

            if (value)
                videoPlayer.Play();
            else 
                videoPlayer.Pause();
        }       

        /// <summary>
        /// Toggle loop video clip.
        /// </summary>
        /// <param name="value"></param>
        private void LoopToggle(bool value)
        {
            videoPlayer.isLooping = value;
            PlayToggle(isPlaying.val); //start play immediatly if play is checked
        }

        /// <summary>
        /// Opens dialog for select video clip.
        /// </summary>
        private void OpenVideoClipBrowser()
        {
            SuperController.singleton.NormalizeMediaPath(DefaultVideoPath);
            SuperController.singleton.GetMediaPathDialog(LoadVideoClip);
        }

        /// <summary>
        /// Adds required componets to target GameObject.
        /// </summary>
        public void AddVideoComponents(string targetObject)
        {
            OnDestroy();

            var sourceObject = containingAtom.containingSubScene == null ? containingAtom.gameObject : containingAtom.parentAtom.gameObject;            

            var screen = GetImageGameObject(sourceObject, targetObject);            

            if (screen == null)
                throw new Exception("Unable to find object for render video");  

            videoPlayer = screen.AddComponent<VideoPlayer>();
            audioSource = screen.AddComponent<AudioSource>();

            videoPlayer.playOnAwake = false;
            audioSource.playOnAwake = false;

            videoPlayer.renderMode = VideoRenderMode.MaterialOverride;
            videoPlayer.audioOutputMode = VideoAudioOutputMode.AudioSource;
            videoPlayer.controlledAudioTrackCount = 1;
            videoPlayer.EnableAudioTrack(0, true);
            videoPlayer.SetTargetAudioSource(0, audioSource);
            videoPlayer.isLooping = isLooping.val;

            audioSource.volume = audioVolume.val;
            audioSource.spatialBlend = audioSpatialBlend.val;
            audioSource.spatialize = audioSpatialize.val;
            audioSource.spread = audioSpread.val;
            audioSource.minDistance = audioMinDistance.val;
            audioSource.maxDistance = audioMaxDistance.val;

            LoadVideoClip(videoClipPath.val);
        }

        /// <summary>
        /// Loads video clip.
        /// </summary>
        /// <param name="path">Video clip absolute path</param>
        private void LoadVideoClip(string path)
        {
            if (string.IsNullOrEmpty(path)) return;

            videoPlayer.Stop();
            videoClipPath.val = path;
            chooseVideoBtn.buttonText.text = path;
            chooseVideoBtn.buttonColor = new Color(0, 255, 0, 0.5f);
            videoPlayer.url = videoClipPath.val;
            PlayToggle(isPlaying.val);
        }
        #endregion

        /// <summary>
        /// Create script UI.
        /// </summary>
        private void CreateUI()
        {
            CreateScriptCaptionUI(new JSONStorableString("scriptCaption",
                "<b><size=40>RenderVideo</size></b>\n" +
                "Plays video on various VAM devices and shapes"));            

            CreateSpacer(true).height = 120f;

            CreateGroupCaptionUI(new JSONStorableString("videoSettingsCaption","<b>Video Settings</b>"));
            CreateGroupCaptionUI(new JSONStorableString("audioSettingsCaption","<b>Audio Settings</b>"), true);

            chooseVideoBtn = CreateButton("Choose video clip...");
            chooseVideoBtn.button.onClick.AddListener(OpenVideoClipBrowser);
            chooseVideoBtn.height = 120f;
            chooseVideoBtn.buttonColor = new Color(255, 255, 0, 0.5f);

            CreatePopup(videoObject);
            CreateToggle(isPlaying);
            CreateToggle(isLooping);
            CreateSlider(audioVolume, true);
            CreateSlider(audioSpatialBlend, true);
            CreateToggle(audioSpatialize, true);
            CreateSlider(audioSpread, true);
            CreateSlider(audioMinDistance, true);
            CreateSlider(audioMaxDistance, true);            
        }

        /// <summary>
        /// Initialize and register script storables.
        /// </summary>
        private void RegisterStorables()
        {
            var videoObjectList = supportedAtoms[containingAtom.type];
            videoObject = new JSONStorableStringChooser("Video Object", videoObjectList, videoObjectList.First(), "Screen", AddVideoComponents);
            RegisterStringChooser(videoObject);

            videoClipPath = new JSONStorableString("Video Clip Path", string.Empty, LoadVideoClip);
            RegisterString(videoClipPath);

            isPlaying = new JSONStorableBool("Play", true, PlayToggle);
            RegisterBool(isPlaying);

            isLooping = new JSONStorableBool("Loop", true, LoopToggle);
            RegisterBool(isLooping);

            audioSpatialize = new JSONStorableBool("Spatialize", true, (val) => audioSource.spatialize = val);
            RegisterBool(audioSpatialize);

            audioVolume = new JSONStorableFloat("Audio Volume", 1, (val) => audioSource.volume = val, 0, 1);
            RegisterFloat(audioVolume);

            audioSpatialBlend = new JSONStorableFloat("Spatial Blend", 1, (val) => audioSource.spatialBlend = val, 0, 1);
            RegisterFloat(audioSpatialBlend);

            audioSpread = new JSONStorableFloat("Spread", 0, (val) => audioSource.spread = val, 0, 360);
            RegisterFloat(audioSpread);

            audioMinDistance = new JSONStorableFloat("Min Distance", 1, (val) => audioSource.minDistance = val, 0, 500);
            RegisterFloat(audioMinDistance);

            audioMaxDistance = new JSONStorableFloat("Max Distance", 500, (val) => audioSource.maxDistance = val, 0, 500);
            RegisterFloat(audioMaxDistance);
        }
       
        /// <summary>
        /// Init script procedure
        /// </summary>
        public override void Init()
        {
            try
            {                               
                // check if current atom supported.
                if (!supportedAtoms.ContainsKey(containingAtom.type))
                    throw new NotImplementedException($"{containingAtom.type} not supported. Supported atoms: {string.Join(",",supportedAtoms.Keys.ToList().ToArray())}");                                
                
                // register storables
                RegisterStorables();

                // create script UI
                CreateUI();

                // add video components 
                AddVideoComponents(videoObject.val);

                // load clip
                LoadVideoClip(videoClipPath.val);
            }
            catch (NotImplementedException e)
            {
                SuperController.LogError($"{this.name} : {e.Message}");
            }
            catch (Exception e)
            {
                SuperController.LogError($"{this.name} : Unhandled exception: " + e);
            }
        }

        /// <summary>
        /// Mute sound freeze motion/sound button is checked.
        /// </summary>
        void Update()
        {                        
            if (SuperController.singleton.freezeAnimation)
                audioSource.mute = true;            
            else 
                audioSource.mute = false;
        }   

        /// <summary>
        /// Destroy created components on exit/reload
        /// </summary>
        void OnDestroy()
        {
            if (videoPlayer != null) Destroy(videoPlayer);
            if (audioSource != null) Destroy(audioSource);
        }
    }
}
